/**
 * Copyright (c) Nicolas Gallagher
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */

/**
 * Hook for integrating the Responder System into React
 *
 *   function SomeComponent({ onStartShouldSetResponder }) {
 *     const ref = useRef(null);
 *     useResponderEvents(ref, { onStartShouldSetResponder });
 *     return <div ref={ref} />
 *   }
 */
import * as React from 'react';
import * as ResponderSystem from './ResponderSystem';
var emptyObject = {};
var idCounter = 0;

function useStable(getInitialValue) {
  var ref = React.useRef(null);

  if (ref.current == null) {
    ref.current = getInitialValue();
  }

  return ref.current;
}

export default function useResponderEvents(hostRef, callbacks) {
  if (callbacks === void 0) {
    callbacks = emptyObject;
  }

  var id = useStable(function () {
    return idCounter++;
  });
  var isAttachedRef = React.useRef(false); // These are separate effects so they doesn't run when `callbacks` changes.
  // On initial mount, attach global listeners if needed.

  React.useEffect(function () {
    ResponderSystem.attachListeners();
  }, []); // On unmount, remove node potentially attached to the Responder System.

  React.useEffect(function () {
    return function () {
      ResponderSystem.removeNode(id);
    };
  }, [id]); // Register and unregister with the Responder System as necessary

  React.useEffect(function () {
    var _callbacks = callbacks,
        onMoveShouldSetResponder = _callbacks.onMoveShouldSetResponder,
        onMoveShouldSetResponderCapture = _callbacks.onMoveShouldSetResponderCapture,
        onScrollShouldSetResponder = _callbacks.onScrollShouldSetResponder,
        onScrollShouldSetResponderCapture = _callbacks.onScrollShouldSetResponderCapture,
        onSelectionChangeShouldSetResponder = _callbacks.onSelectionChangeShouldSetResponder,
        onSelectionChangeShouldSetResponderCapture = _callbacks.onSelectionChangeShouldSetResponderCapture,
        onStartShouldSetResponder = _callbacks.onStartShouldSetResponder,
        onStartShouldSetResponderCapture = _callbacks.onStartShouldSetResponderCapture;
    var requiresResponderSystem = onMoveShouldSetResponder != null || onMoveShouldSetResponderCapture != null || onScrollShouldSetResponder != null || onScrollShouldSetResponderCapture != null || onSelectionChangeShouldSetResponder != null || onSelectionChangeShouldSetResponderCapture != null || onStartShouldSetResponder != null || onStartShouldSetResponderCapture != null;
    var node = hostRef.current;

    if (requiresResponderSystem) {
      ResponderSystem.addNode(id, node, callbacks);
      isAttachedRef.current = true;
    } else if (isAttachedRef.current) {
      ResponderSystem.removeNode(id);
      isAttachedRef.current = false;
    }
  }, [callbacks, hostRef, id]);
}