/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow strict-local
 * @format
 */

'use strict';

import type { PressabilityConfig } from './Pressability';
import type { ViewProps } from '../View';

import * as React from 'react';
import { useMemo, useState, useRef, useImperativeHandle } from 'react';
import usePressability from './usePressability';
import View from '../View';

export type StateCallbackType = $ReadOnly<{|
  focused: boolean,
  pressed: boolean
|}>;

type ViewStyleProp = $PropertyType<ViewProps, 'style'>;

type Props = $ReadOnly<{|
  accessibilityLabel?: $PropertyType<ViewProps, 'accessibilityLabel'>,
  accessibilityLiveRegion?: $PropertyType<ViewProps, 'accessibilityLiveRegion'>,
  accessibilityRole?: $PropertyType<ViewProps, 'accessibilityRole'>,
  accessibilityState?: $PropertyType<ViewProps, 'accessibilityState'>,
  accessibilityValue?: $PropertyType<ViewProps, 'accessibilityValue'>,
  accessible?: $PropertyType<ViewProps, 'accessible'>,
  focusable?: ?boolean,
  importantForAccessibility?: $PropertyType<ViewProps, 'importantForAccessibility'>,
  /**
   * Either children or a render prop that receives a boolean reflecting whether
   * the component is currently pressed.
   */
  children: React.Node | ((state: StateCallbackType) => React.Node),
  /**
   * Duration (in milliseconds) from `onPressIn` before `onLongPress` is called.
   */
  delayLongPress?: ?number,
  /**
   * Whether the press behavior is disabled.
   */
  disabled?: ?boolean,
  /**
   * Additional distance outside of this view in which a press is detected.
   */
  hitSlop?: $PropertyType<ViewProps, 'hitSlop'>,
  /**
   * Called when...
   */
  onBlur?: $PropertyType<PressabilityConfig, 'onBlur'>,
  /**
   * Called when...
   */
  onFocus?: $PropertyType<PressabilityConfig, 'onFocus'>,
  /**
   * Called when this view's layout changes.
   */
  onLayout?: $PropertyType<ViewProps, 'onLayout'>,
  /**
   * Called when a long-tap gesture is detected.
   */
  onLongPress?: $PropertyType<PressabilityConfig, 'onLongPress'>,
  /**
   * Called when a single tap gesture is detected.
   */
  onPress?: $PropertyType<PressabilityConfig, 'onPress'>,
  /**
   * Called when a touch is engaged before `onPress`.
   */
  onPressIn?: $PropertyType<PressabilityConfig, 'onPressIn'>,
  /**
   * Called when a touch is engaged before `onPress`.
   */
  onPressMove?: $PropertyType<PressabilityConfig, 'onPressMove'>,
  /**
   * Called when a touch is released before `onPress`.
   */
  onPressOut?: $PropertyType<PressabilityConfig, 'onPressOut'>,
  /**
   * Either view styles or a function that receives a boolean reflecting whether
   * the component is currently pressed and returns view styles.
   */
  style?: ViewStyleProp | ((state: StateCallbackType) => ViewStyleProp),
  /**
   * Identifier used to find this view in tests.
   */
  testID?: $PropertyType<ViewProps, 'testID'>,
  /**
   * Used only for documentation or testing (e.g. snapshot testing).
   */
  testOnly_pressed?: ?boolean
|}>;

/**
 * Component used to build display components that should respond to whether the
 * component is currently pressed or not.
 */
function Pressable(props: Props, forwardedRef): React.Node {
  const {
    accessible,
    children,
    delayLongPress,
    disabled,
    focusable,
    hitSlop,
    onBlur,
    onFocus,
    onLongPress,
    onPress,
    onPressMove,
    onPressIn,
    onPressOut,
    style,
    testOnly_pressed,
    ...restProps
  } = props;

  const viewRef = useRef<React.ElementRef<typeof View> | null>(null);
  useImperativeHandle(forwardedRef, () => viewRef.current);

  const [pressed, setPressed] = usePressState(testOnly_pressed === true);
  const [focused, setFocused] = useState(false);

  const config = useMemo(
    () => ({
      disabled,
      hitSlop,
      delayLongPress,
      onBlur(event) {
        setFocused(false);
        if (onBlur != null) {
          onBlur(event);
        }
      },
      onFocus(event) {
        setFocused(true);
        if (onFocus != null) {
          onFocus(event);
        }
      },
      onLongPress,
      onPress,
      onPressIn(event) {
        setPressed(true);
        if (onPressIn != null) {
          onPressIn(event);
        }
      },
      onPressMove(event) {
        if (onPressMove != null) {
          onPressMove(event);
        }
      },
      onPressOut(event) {
        setPressed(false);
        if (onPressOut != null) {
          onPressOut(event);
        }
      }
    }),
    [
      delayLongPress,
      disabled,
      hitSlop,
      onBlur,
      onFocus,
      onLongPress,
      onPress,
      onPressIn,
      onPressMove,
      onPressOut,
      setPressed
    ]
  );
  const eventHandlers = usePressability(config);

  return (
    <View
      {...restProps}
      accessibilityRole={restProps.accessibilityRole ?? 'button'}
      accessibilityState={{
        disabled: this.props.disabled,
        ...this.props.accessibilityState
      }}
      accessible={accessible !== false}
      focusable={focusable !== false}
      hitSlop={hitSlop}
      ref={viewRef}
      style={typeof style === 'function' ? style({ focused, pressed }) : style}
      {...eventHandlers}
    >
      {typeof children === 'function' ? children({ focused, pressed }) : children}
    </View>
  );
}

function usePressState(forcePressed: boolean): [boolean, (boolean) => void] {
  const [pressed, setPressed] = useState(false);
  return [pressed || forcePressed, setPressed];
}

const MemoedPressable = React.memo(React.forwardRef(Pressable));
MemoedPressable.displayName = 'Pressable';

export default (MemoedPressable: React.AbstractComponent<Props, React.ElementRef<typeof View>>);
