/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow strict-local
 * @format
 */

'use strict';

import type { ColorValue } from '../../types';
import type { PressabilityConfig } from '../Pressable/Pressability';
import type { Props as TouchableWithoutFeedbackProps } from '../TouchableWithoutFeedback';
import type { ViewProps } from '../View';

import * as React from 'react';
import Pressability from '../Pressable/Pressability';
import StyleSheet from '../StyleSheet';
import View from '../View';

type ViewStyle = $PropertyType<ViewProps, 'style'>;

type Props = $ReadOnly<{|
  ...TouchableWithoutFeedbackProps,
  activeOpacity?: ?number,
  underlayColor?: ?ColorValue,
  onShowUnderlay?: ?() => void,
  onHideUnderlay?: ?() => void,
  style?: ViewStyle,
  testOnly_pressed?: ?boolean,

  hostRef: React.Ref<typeof View>
|}>;

type ExtraStyles = $ReadOnly<{|
  child: ViewStyle,
  underlay: ViewStyle
|}>;

type State = $ReadOnly<{|
  pressability: Pressability,
  extraStyles: ?ExtraStyles
|}>;

/**
 * A wrapper for making views respond properly to touches.
 * On press down, the opacity of the wrapped view is decreased, which allows
 * the underlay color to show through, darkening or tinting the view.
 *
 * The underlay comes from wrapping the child in a new View, which can affect
 * layout, and sometimes cause unwanted visual artifacts if not used correctly,
 * for example if the backgroundColor of the wrapped view isn't explicitly set
 * to an opaque color.
 *
 * TouchableHighlight must have one child (not zero or more than one).
 * If you wish to have several child components, wrap them in a View.
 */
class TouchableHighlight extends React.Component<Props, State> {
  _hideTimeout: ?TimeoutID;
  _isMounted: boolean = false;

  state: State = {
    pressability: new Pressability(this._createPressabilityConfig()),
    extraStyles: this.props.testOnly_pressed === true ? this._createExtraStyles() : null
  };

  _createPressabilityConfig(): PressabilityConfig {
    return {
      cancelable: !this.props.rejectResponderTermination,
      disabled: this.props.disabled,
      hitSlop: this.props.hitSlop,
      delayLongPress: this.props.delayLongPress,
      delayPressIn: this.props.delayPressIn,
      delayPressOut: this.props.delayPressOut,
      onBlur: event => {
        if (this.props.onBlur != null) {
          this.props.onBlur(event);
        }
      },
      onFocus: event => {
        if (this.props.onFocus != null) {
          this.props.onFocus(event);
        }
      },
      onLongPress: event => {
        if (this.props.onLongPress != null) {
          this.props.onLongPress(event);
        }
      },
      onPress: event => {
        if (this._hideTimeout != null) {
          clearTimeout(this._hideTimeout);
        }
        if (this.props.onPress != null) {
          this.props.onPress(event);
        }
      },
      onPressIn: event => {
        if (this._hideTimeout != null) {
          clearTimeout(this._hideTimeout);
          this._hideTimeout = null;
        }
        this._showUnderlay();
        if (this.props.onPressIn != null) {
          this.props.onPressIn(event);
        }
      },
      onPressOut: event => {
        if (this._hideTimeout == null) {
          this._hideUnderlay();
        }
        if (this.props.onPressOut != null) {
          this.props.onPressOut(event);
        }
      }
    };
  }

  _createExtraStyles(): ExtraStyles {
    return {
      child: { opacity: this.props.activeOpacity ?? 0.85 },
      underlay: {
        backgroundColor: this.props.underlayColor === undefined ? 'black' : this.props.underlayColor
      }
    };
  }

  _showUnderlay(): void {
    if (!this._isMounted || !this._hasPressHandler()) {
      return;
    }
    this.setState(() => ({ extraStyles: this._createExtraStyles() }));
    if (this.props.onShowUnderlay != null) {
      this.props.onShowUnderlay();
    }
  }

  _hideUnderlay(): void {
    if (this._hideTimeout != null) {
      clearTimeout(this._hideTimeout);
      this._hideTimeout = null;
    }
    if (this.props.testOnly_pressed === true) {
      return;
    }
    if (this._hasPressHandler()) {
      this.setState(() => ({ extraStyles: null }));
      if (this.props.onHideUnderlay != null) {
        this.props.onHideUnderlay();
      }
    }
  }

  _hasPressHandler(): boolean {
    return (
      this.props.onPress != null ||
      this.props.onPressIn != null ||
      this.props.onPressOut != null ||
      this.props.onLongPress != null
    );
  }

  render(): React.Node {
    const child = React.Children.only(this.props.children);
    const eventHandlers = this.state.pressability.getEventHandlers();

    return (
      <View
        accessibilityLabel={this.props.accessibilityLabel}
        accessibilityLiveRegion={this.props.accessibilityLiveRegion}
        accessibilityRole={this.props.accessibilityRole}
        accessibilityState={{
          disabled: this.props.disabled,
          ...this.props.accessibilityState
        }}
        accessibilityValue={this.props.accessibilityValue}
        accessible={this.props.accessible !== false}
        focusable={this.props.focusable !== false && this.props.onPress !== undefined}
        hitSlop={this.props.hitSlop}
        importantForAccessibility={this.props.importantForAccessibility}
        nativeID={this.props.nativeID}
        onLayout={this.props.onLayout}
        ref={this.props.hostRef}
        style={[
          styles.root,
          this.props.style,
          !this.props.disabled && styles.actionable,
          this.state.extraStyles && this.state.extraStyles.underlay
        ]}
        testID={this.props.testID}
        {...eventHandlers}
      >
        {React.cloneElement(child, {
          style: StyleSheet.compose(
            child.props.style,
            this.state.extraStyles && this.state.extraStyles.child
          )
        })}
      </View>
    );
  }

  componentDidMount(): void {
    this._isMounted = true;
  }

  componentDidUpdate(prevProps: Props, prevState: State) {
    this.state.pressability.configure(this._createPressabilityConfig());
  }

  componentWillUnmount(): void {
    this._isMounted = false;
    if (this._hideTimeout != null) {
      clearTimeout(this._hideTimeout);
    }
    this.state.pressability.reset();
  }
}

const TouchableHighlightWrapper = (React.forwardRef((props, hostRef) => (
  <TouchableHighlight {...props} hostRef={hostRef} />
)): React.ComponentType<$ReadOnly<$Diff<Props, {| hostRef: mixed |}>>>);

TouchableHighlightWrapper.displayName = 'TouchableHighlight';

export default TouchableHighlightWrapper;

const styles = StyleSheet.create({
  root: {
    userSelect: 'none'
  },
  actionable: {
    cursor: 'pointer',
    touchAction: 'manipulation'
  }
});
