/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow strict-local
 * @format
 */

'use strict';

import type { PressabilityConfig } from '../Pressable/Pressability';
import type { Props as TouchableWithoutFeedbackProps } from '../TouchableWithoutFeedback';
import type { ViewProps } from '../View';

import * as React from 'react';
import Pressability from '../Pressable/Pressability';
import StyleSheet from '../StyleSheet';
import View from '../View';

type ViewStyle = $PropertyType<ViewProps, 'style'>;

type Props = $ReadOnly<{|
  ...TouchableWithoutFeedbackProps,
  activeOpacity?: ?number,
  style?: ?ViewStyle,
  hostRef: React.Ref<typeof View>
|}>;

type State = $ReadOnly<{|
  duration: string,
  opacity: number,
  pressability: Pressability
|}>;

/**
 * A wrapper for making views respond properly to touches.
 * On press down, the opacity of the wrapped view is decreased, dimming it.
 *
 * Opacity is controlled by wrapping the children in an Animated.View, which is
 * added to the view hierarchy.  Be aware that this can affect layout.
 */
class TouchableOpacity extends React.Component<Props, State> {
  state: State = {
    duration: '0s',
    opacity: this._getChildStyleOpacityWithDefault(),
    pressability: new Pressability(this._createPressabilityConfig())
  };

  _createPressabilityConfig(): PressabilityConfig {
    return {
      cancelable: !this.props.rejectResponderTermination,
      disabled: this.props.disabled,
      hitSlop: this.props.hitSlop,
      delayLongPress: this.props.delayLongPress,
      delayPressIn: this.props.delayPressIn,
      delayPressOut: this.props.delayPressOut,
      onBlur: event => {
        if (this.props.onBlur != null) {
          this.props.onBlur(event);
        }
      },
      onFocus: event => {
        if (this.props.onFocus != null) {
          this.props.onFocus(event);
        }
      },
      onLongPress: this.props.onLongPress,
      onPress: this.props.onPress,
      onPressIn: event => {
        this._opacityActive(event.dispatchConfig.registrationName === 'onResponderGrant' ? 0 : 150);
        if (this.props.onPressIn != null) {
          this.props.onPressIn(event);
        }
      },
      onPressOut: event => {
        this._opacityInactive(250);
        if (this.props.onPressOut != null) {
          this.props.onPressOut(event);
        }
      }
    };
  }

  /**
   * Animate the touchable to a new opacity.
   */
  _setOpacityTo(value: number, duration: number): void {
    this.setState(() => ({
      opacity: value,
      duration: duration ? `${duration / 1000}s` : '0s'
    }));
  }

  _opacityActive(duration: number): void {
    this._setOpacityTo(this.props.activeOpacity ?? 0.2, duration);
  }

  _opacityInactive(duration: number): void {
    this._setOpacityTo(this._getChildStyleOpacityWithDefault(), duration);
  }

  _getChildStyleOpacityWithDefault(): number {
    const flatStyle = StyleSheet.flatten(this.props.style);
    const opacity = flatStyle != null ? flatStyle.opacity : null;
    return typeof opacity === 'number' ? opacity : 1;
  }

  render(): React.Node {
    const eventHandlers = this.state.pressability.getEventHandlers();

    return (
      <View
        accessibilityLabel={this.props.accessibilityLabel}
        accessibilityLiveRegion={this.props.accessibilityLiveRegion}
        accessibilityRole={this.props.accessibilityRole}
        accessibilityState={{
          disabled: this.props.disabled,
          ...this.props.accessibilityState
        }}
        accessibilityValue={this.props.accessibilityValue}
        accessible={this.props.accessible !== false}
        focusable={this.props.focusable !== false && this.props.onPress !== undefined}
        hitSlop={this.props.hitSlop}
        importantForAccessibility={this.props.importantForAccessibility}
        nativeID={this.props.nativeID}
        onLayout={this.props.onLayout}
        ref={this.props.hostRef}
        style={[
          styles.root,
          !this.props.disabled && styles.actionable,
          this.props.style,
          {
            opacity: this.state.opacity,
            transitionDuration: this.state.duration
          }
        ]}
        testID={this.props.testID}
        {...eventHandlers}
      >
        {this.props.children}
      </View>
    );
  }

  componentDidUpdate(prevProps: Props, prevState: State) {
    this.state.pressability.configure(this._createPressabilityConfig());
    if (this.props.disabled !== prevProps.disabled) {
      this._opacityInactive(250);
    }
  }

  componentWillUnmount(): void {
    this.state.pressability.reset();
  }
}

const TouchableOpacityWrapper = (React.forwardRef((props, hostRef) => (
  <TouchableOpacity {...props} hostRef={hostRef} />
)): React.ComponentType<$ReadOnly<$Diff<Props, {| hostRef: mixed |}>>>);

TouchableOpacityWrapper.displayName = 'TouchableOpacity';

const styles = StyleSheet.create({
  root: {
    transitionProperty: 'opacity',
    transitionDuration: '0.15s',
    userSelect: 'none'
  },
  actionable: {
    cursor: 'pointer',
    touchAction: 'manipulation'
  }
});

export default TouchableOpacityWrapper;
